#version 330
//definice funkcí pro výpočet vzdálenosti od objektu

float sdPlane(vec3 point, vec4 n) {
    //n musí být normalizované
    return dot(point, n.xyz) + n.w;
}
float sdTorus(vec3 p, vec2 t)
{
    vec2 q = vec2(length(p.xy)-t.x, p.z);
    return length(q)-t.y;
}
float sdVerticalCapsule(vec3 p, float h, float r)
{
    p.y -= clamp(p.y, 0.0, h);
    return length(p) - r;
}

float sdBox(vec3 point) {
    vec3 d = abs(point)-1;
    return min(max(d.x, max(d.y, d.z)), 0.0) + length(max(d, 0.0));
}

float sdSphere(vec3 point) {
    return length(point)-1;
}

float sdTriPrism(vec3 point, vec2 h) {
    vec3 q = abs(point);
    return max(q.z-h.y, max(q.x*0.866025+point.y*0.5, -point.y)-h.x*0.5);
}

float calculateDisplacement(vec3 point, float displacement) {
    return sin(displacement * point.x) * sin(displacement * point.y) * sin(displacement * point.z) * 0.25;
}
const float Power = 3.0;
float SDF(vec3 pos, out int steps) {
    vec3 z = pos;
    float dr = 1.0;
    float r = 0.0;
    for (int i = 0; i < 20; i++) {
        r = length(z);
        steps = i;
        if (r>4.0) break;

        // convert to polar coordinates
        float theta = acos(z.z/r);
        float phi = atan(z.y, z.x);
        dr =  pow(r, Power-1.0)*Power*dr + 1.0;

        // scale and rotate the point
        float zr = pow(r, Power);
        theta = theta*Power;
        phi = phi*Power;

        // convert back to cartesian coordinates
        z = zr*vec3(sin(theta)*cos(phi), sin(phi)*sin(theta), cos(theta));
        z+=pos;
    }

    return 0.5*log(r)*r/dr;
}

struct Hit {
    float distance;
    vec3 col;
    float reflectivity;
    float refractivity;
};
//transformace
vec3 translate(vec3 p, vec3 t) {
    return p - t;
}
vec3 scale(vec3 p, vec3 s) {
    mat4 S = mat4(
    vec4(1/s.x, 0, 0, 0),
    vec4(0, 1/s.y, 0, 0),
    vec4(0, 0, 1/s.z, 0),
    vec4(0, 0, 0, 1));
    return (vec4(p, 1)*S).xyz;
}
vec3 rot3D(vec3 p, vec3 axis, float angle) {
    angle = radians(angle);
    return mix(dot(axis, p) * axis, p, cos(angle)) + cross(axis, p) * sin(angle);
}

//definice operátorů pro kombinování objektů
float sdUnion(float o1, float o2) {
    return (o1 < o2) ? o1 : o2;
}
float sdSubstraction(float o1, float o2) {
    o1 = -o1;
    return (o1 > o2) ? o1 : o2;
}
float sdIntersection(float o1, float o2) {
    return (o1 > o2) ? o1 : o2;
}
float sdUnion(float o1, float o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) - k*h*(1.0-h);

}
float sdSubstraction(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2+o1)/k, 0.0, 1.0);
    return mix(o2, -o1, h) + k*h*(1.0-h);
}

float sdIntersection(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) + k*h*(1.0-h);
}
Hit sdUnion(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) - k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);
    float refractivity = mix(o2.refractivity, o1.refractivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;
    obj.refractivity = refractivity;

    return obj;
}
Hit sdSubstraction(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance+o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, -o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}

Hit sdIntersection(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}
//modulo funkce
float pMod (inout float p, float size)
{
    float halfsize = size * 0.5;
    float c = floor((p+halfsize)/size);
    p = mod(p+halfsize, size)-halfsize;
    p = mod(-p+halfsize, size)-halfsize;
    return c;
}
vec3 opTwistY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opTwistX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opTwistZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}

vec3 opBendX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}
vec3 opBendY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opBendZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opSymX(vec3 p)
{
    p.x = abs(p.x);
    return p;
}
vec3 opSymY(vec3 p)
{
    p.y = abs(p.y);
    return p;
}
vec3 opSymZ(vec3 p)
{
    p.z = abs(p.z);
    return p;
}
vec3 opSymXY(vec3 p)
{
    p.xy = abs(p.xy);
    return p;
}
vec3 opSymXZ(vec3 p)
{
    p.xz = abs(p.xz);
    return p;
}
vec3 opSymYZ(vec3 p)
{
    p.yz = abs(p.yz);
    return p;
}
vec3 opSymXYZ(vec3 p)
{
    p = abs(p);
    return p;
}
vec3 repeated( vec3 p, float s )
{
    vec3 r = p - s*round(p/s);
    return r;
}
//vec3 opRepX( vec3 p, float k )
//{
//    return mod(p,k)-k*0.5;
//}
vec3 refract(vec3 rayDir, vec3 normal, float n1, float n2) {
    float n = n1/n2;
    float cosI = -dot(normal, rayDir);
    float sinT2 = n*n*(1.0-cosI*cosI);
    if (sinT2 > 1.0) {
        return vec3(0.0);
    }
    float cosT = sqrt(1.0-sinT2);
    return n*rayDir + (n*cosI-cosT)*normal;
}
vec3 reflect(vec3 rayDir, vec3 normal) {
    return rayDir - 2.0*dot(rayDir, normal)*normal;
}

////////////////////////////////////////////////////////////////////////


in vec3 vertColor;
out vec4 outColor;

uniform float width;
uniform float height;
//proměná, která se mění s časem
uniform vec3 camPos;
uniform mat4x4 lookAt;
uniform float time;

const int steps = 1000;
const float minDist = 0.001;
const float maxDist = 100.0;

layout(origin_upper_left) in vec4 gl_FragCoord;


float mapScene(vec3 point) {

        float sphere = sdSphere(point);
        float box = sdBox(translate(opSymX(opSymY(opSymZ(point))), vec3(time, time, time)));

        float scene = sdUnion(box, sphere, 10);
        return scene;
}

float rayMarching(vec3 rayOri, vec3 rayDir) {
    float distTraveled = 0.0;
    vec3 currentPos = rayOri;

    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }
        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně
        float distToClosest = mapScene(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {

            //                        return shading(currentPos, normal, getMaterial(currentPos, Hit.y, normal, reflective));
            //                        return Hit.col;
            return distTraveled;
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return distTraveled;
}


void main() {
    //řešení velikosti okna a poměru stran
    vec2 uv = vec2(((gl_FragCoord.x - 0.5) / width)*2-1, ((gl_FragCoord.y - 0.5) / height)*2-1);
    float aspectRatio = width/height;
    if (aspectRatio < 1) {
        uv.y *= 1/aspectRatio;
    } else {
        uv.x *= aspectRatio;
    }
    //počáteční pozice paprsku
    vec3 rayOri = camPos;
    //směr paprsku
    vec4 temp =vec4(uv, 1, 1)*lookAt;
    vec3 rayDir = normalize(temp.xyz);
    //výsledná barva pixelu
    vec3 color = vec3(1-rayMarching(rayOri, rayDir)*0.05);
    outColor = vec4(color, 1.0);
}

