#version 330
//definice funkcí pro výpočet vzdálenosti od objektu

float sdPlane(vec3 point, vec4 n) {
    //n musí být normalizované
    return dot(point, n.xyz) + n.w;
}
float sdTorus(vec3 p, vec2 t)
{
    vec2 q = vec2(length(p.xy)-t.x, p.z);
    return length(q)-t.y;
}
float sdVerticalCapsule(vec3 p, float h, float r)
{
    p.y -= clamp(p.y, 0.0, h);
    return length(p) - r;
}

float sdBox(vec3 point) {
    vec3 d = abs(point)-1;
    return min(max(d.x, max(d.y, d.z)), 0.0) + length(max(d, 0.0));
}

float sdSphere(vec3 point) {
    return length(point)-1;
}

float sdTriPrism(vec3 point, vec2 h) {
    vec3 q = abs(point);
    return max(q.z-h.y, max(q.x*0.866025+point.y*0.5, -point.y)-h.x*0.5);
}

float calculateDisplacement(vec3 point, float displacement) {
    return sin(displacement * point.x) * sin(displacement * point.y) * sin(displacement * point.z) * 0.25;
}

struct Hit {
    float distance;
    vec3 col;
    float reflectivity;
    float refractivity;
};
//transformace
vec3 translate(vec3 p, vec3 t) {
    return p - t;
}
vec3 scale(vec3 p, vec3 s) {
    mat4 S = mat4(
    vec4(1/s.x, 0, 0, 0),
    vec4(0, 1/s.y, 0, 0),
    vec4(0, 0, 1/s.z, 0),
    vec4(0, 0, 0, 1));
    return (vec4(p, 1)*S).xyz;
}
vec3 rot3D(vec3 p, vec3 axis, float angle) {
    angle = radians(angle);
    return mix(dot(axis, p) * axis, p, cos(angle)) + cross(axis, p) * sin(angle);
}

//definice operátorů pro kombinování objektů
float sdUnion(float o1, float o2) {
    return (o1 < o2) ? o1 : o2;
}
float sdSubstraction(float o1, float o2) {
    o1 = -o1;
    return (o1 > o2) ? o1 : o2;
}
float sdIntersection(float o1, float o2) {
    return (o1 > o2) ? o1 : o2;
}
float sdUnion(float o1, float o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) - k*h*(1.0-h);

}
float sdSubtraction(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2+o1)/k, 0.0, 1.0);
    return mix(o2, -o1, h) + k*h*(1.0-h);
}

float sdIntersection(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) + k*h*(1.0-h);
}
Hit sdUnion(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) - k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);
    float refractivity = mix(o2.refractivity, o1.refractivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;
    obj.refractivity = refractivity;

    return obj;
}
Hit sdSubtraction(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance+o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, -o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}
Hit sdIntersection(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}
//transformační funkce
//modulo funkce
float pMod (inout float p, float size)
{
    float halfsize = size * 0.5;
    float c = floor((p+halfsize)/size);
    p = mod(p+halfsize, size)-halfsize;
    p = mod(-p+halfsize, size)-halfsize;
    return c;
}
vec3 opTwistY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opTwistX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opTwistZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}

vec3 opBendX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}
vec3 opBendY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opBendZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opSymX(vec3 p)
{
    p.x = abs(p.x);
    return p;
}
vec3 opSymY(vec3 p)
{
    p.y = abs(p.y);
    return p;
}
vec3 opSymZ(vec3 p)
{
    p.z = abs(p.z);
    return p;
}
vec3 opSymXY(vec3 p)
{
    p.xy = abs(p.xy);
    return p;
}
vec3 opSymXZ(vec3 p)
{
    p.xz = abs(p.xz);
    return p;
}
vec3 opSymYZ(vec3 p)
{
    p.yz = abs(p.yz);
    return p;
}
vec3 opSymXYZ(vec3 p)
{
    p = abs(p);
    return p;
}
vec3 repeated( vec3 p, float s )
{
    vec3 r = p - s*round(p/s);
    return r;
}
//vec3 opRepX( vec3 p, float k )
//{
//    return mod(p,k)-k*0.5;
//}
vec3 refract(vec3 rayDir, vec3 normal, float n1, float n2) {
    float n = n1/n2;
    float cosI = -dot(normal, rayDir);
    float sinT2 = n*n*(1.0-cosI*cosI);
    if (sinT2 > 1.0) {
        return vec3(0.0);
    }
    float cosT = sqrt(1.0-sinT2);
    return n*rayDir + (n*cosI-cosT)*normal;
}
vec3 reflect(vec3 rayDir, vec3 normal) {
    return rayDir - 2.0*dot(rayDir, normal)*normal;
}

////////////////////////////////////////////////////////////////////////


in vec3 vertColor;
out vec4 outColor;

uniform float width;
uniform float height;
//proměná, která se mění s časem
uniform float h;
const int maxNumOfLights = 3;
uniform vec3 camPos;
uniform mat4x4 lookAt;
uniform vec3[maxNumOfLights] lightPos;
uniform vec3[maxNumOfLights] lightCol;
uniform int numOfLights;
uniform sampler2D texture1;
uniform sampler2D texture2;
uniform sampler2D texture3;
uniform int sceneId;


const int steps = 1000;
const float minDist = 0.001;
const float maxDist = 100.0;

//maximální počet odrazů paprsku
float maxRec = 5;

layout(origin_upper_left) in vec4 gl_FragCoord;
//výpočet textury
vec3 calculateTexture(sampler2D texture, vec3 point, vec3 normal) {
    vec3 colorXY = texture2D(texture, mod(point.xy*.5+.5, 1)).xyz;
    vec3 colorYZ = texture2D(texture, mod(point.yz*.5+.5, 1)).xyz;
    vec3 colorXZ = texture2D(texture, mod(point.xz*.5+.5, 1)).xyz;
    normal = abs(normal);
    normal = normal*normal*normal;
    return colorXY*normal.z + colorYZ*normal.x +colorXZ*normal.y;
}

//výpočet fraktálu
vec3 calculateFractal1(vec3 p) {
    float box = sdBox(p);
    vec3 color = vec3( box, 1.0, 0.0 );

    float s = 1.0;
    for( int i=0; i<6; i++ )
    {
        vec3 a = mod( p*s, 2.0 )-1.0;
        s *= 3.0;
        vec3 r = abs(1.0 - 3.0*abs(a));

        float boxa = max(r.x,r.y);
        float boxb = max(r.y,r.z);
        float boxc = max(r.z,r.x);
        float uni = (min(boxa,min(boxb,boxc))-1.0)/s;

        if( uni>box )
        {
            box = uni;
            color = vec3( box, 0.2*boxa*boxb*boxc, (1.0+float(i))/4.0 );
        }
    }

    return color;
}

//výpočet fraktálu zjednodušeně
float preCalculateFractal1(vec3 p) {
    float d = sdBox(p);

    float s = 1.0;
    for( int m=0; m<5; m++ )
    {
        vec3 a = mod( p*s, 2.0 )-1.0;
        s *= 3.0;
        vec3 r = abs(1.0 - 3.0*abs(a));

        float da = max(r.x,r.y);
        float db = max(r.y,r.z);
        float dc = max(r.z,r.x);
        float c = (min(da,min(db,dc))-1.0)/s;

        if( c>d )
        {
            d = c;
        }
    }

    return d;

}

//vytvoření scény
Hit mapScene(vec3 point, vec3 normal) {
    point = scale(point, vec3(10, 10, 10));
    point = repeated(point, 1.3333);
    vec3 fractal = calculateFractal1(point);
    Hit fractalHit = Hit(fractal.x, fractal, 0, 0);
    return fractalHit;
}
float preMapSceneInverted(vec3 point) {
    return 0.0;
}
float preMapScene(vec3 point) {
    point = scale(point, vec3(10, 10, 10));
    point = repeated(point, 1.3333);
    float fractal = preCalculateFractal1(point);
    return fractal;
}

//výpočet normály
vec3 calculateNormal(vec3 point)
{
    const vec2 epsylon = vec2(1e-3, 0);
    vec3 normal = vec3(
    preMapScene(point+epsylon.xyy) - preMapScene(point-epsylon.xyy),
    preMapScene(point+epsylon.yxy) - preMapScene(point-epsylon.yxy),
    preMapScene(point+epsylon.yyx) - preMapScene(point-epsylon.yyx))/epsylon.x;

    return normalize(normal);
}

float hardShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt)
{
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        t += h;
    }
    return 1.0;
}

float softShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt, float k)
{
    float result = 1.0;
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        result = min(result, k*h/t);
        t += h;
    }
    return result;
}

float ambientOcclusion(vec3 p, vec3 n)
{
    float aoStepSize = 0.2;
    float aoIterations = 3;
    float aoIntensity = 0.25;
    float step = aoStepSize;
    float ao = 0.0;
    float dist;
    for (int i = 1; i <= aoIterations; i++)
    {
        dist = step * i;
        ao += max(0.0, (dist - preMapScene(p + n * dist)) / dist);
    }
    return (1.0 - ao * aoIntensity);
}

Hit shading(vec3 point, vec3 normal, Hit obj, vec3 cameraDir)
{
    vec3 color = obj.col;
    vec3 result;
    float lightIntensity = 5;
    float shadowIntensity = 1;
    vec3 light;
    vec3 lightFinal = vec3(0, 0, 0);
    float shadow;
    vec3 lightColor = vec3(0, 0, 0);
    vec3 dirToLight;
    for (int i = 0; i < numOfLights; i++) {
        dirToLight = normalize(lightPos[i] - point);
        float diffuseStrength = max(0.0, dot(dirToLight, normal));
        //phong
//        vec3 reflected = normalize(reflect(dirToLight, normal));
//        float specularStrength = max(0.0, dot(cameraDir, reflected));
//        specularStrength = pow(specularStrength, 32);
        //blinn-phong
        vec3 halfway = normalize(dirToLight - cameraDir);
        float specularStrength = max(0.0, dot(normal, halfway));
        specularStrength = pow(specularStrength, 32);

        light = lightCol[i] * (diffuseStrength*0.333+specularStrength*0.333) * lightIntensity;
        //Vržené stíny
        shadow = hardShadow(point, dirToLight, 0.1, 8) * 0.5 + 0.5;
//        shadow = softShadow(point, dirToLight, 0.1, 12, 8) * 0.5 + 0.5;
        shadow = max(0.0, pow(shadow, shadowIntensity));
        light *= shadow;
        lightColor += lightCol[i];
        lightFinal += light;
    }
    lightColor /= numOfLights;
    float ambientStrenght = 0.5;
    vec3 lightAmbient = ambientStrenght*0.333*lightColor;
    float ambientOcc = ambientOcclusion(point, normal);
    result = color * ambientOcc;

    result *= lightFinal + lightAmbient;
    obj.col = result;
    return obj;
}


Hit rayMarching(vec3 rayOri, vec3 rayDir, inout vec3 currentPos, out vec3 normal) {
    float distTraveled = 0.0;
    const float overRelaxation = 1.5;

    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }

        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně

        float distToClosest = preMapScene(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {
            normal = calculateNormal(currentPos);
            Hit hit = mapScene(currentPos, normal);
            //            return hit;
            return shading(currentPos, normal, hit, rayDir);
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return Hit(0.0, vec3(0.5), 0.0, 0.0);
}

vec3 getResult(vec3 rayOri, vec3 rayDir) {
    vec3 normal;
    vec3 currentPos;
    Hit hit = rayMarching(rayOri, rayDir, currentPos, normal);
    vec3 color = hit.col;
    return color;
}


void main() {
    //řešení velikosti okna a poměru stran
    vec2 uv = vec2(((gl_FragCoord.x - 0.5) / width)*2-1, ((gl_FragCoord.y - 0.5) / height)*2-1);
    float aspectRatio = width/height;
    if (aspectRatio < 1) {
        uv.y *= 1/aspectRatio;
    } else {
        uv.x *= aspectRatio;
    }
    //počáteční pozice paprsku
    vec3 rayOri = camPos;
    //směr paprsku
    vec4 temp =vec4(uv, 1, 1)*lookAt;
    vec3 rayDir = normalize(temp.xyz);
    //výsledná barva pixelu
    vec3 color = getResult(rayOri, rayDir);
    outColor = vec4(color, 1.0);
}

