#version 330
//definice funkcí pro výpočet vzdálenosti od objektu

float sdPlane(vec3 point, vec4 n) {
    //n musí být normalizované
    return dot(point, n.xyz) + n.w;
}
float sdOctahedron(vec3 p, float s)
{
    p = abs(p);
    return (p.x+p.y+p.z-s)*0.57735027;
}
float sdTorus(vec3 p, vec2 t)
{
    vec2 q = vec2(length(p.xy)-t.x, p.z);
    return length(q)-t.y;
}
float sdVerticalCapsule(vec3 p, float h, float r)
{
    p.y -= clamp(p.y, 0.0, h);
    return length(p) - r;
}

float sdBox(vec3 point) {
    vec3 d = abs(point)-1;
    return min(max(d.x, max(d.y, d.z)), 0.0) + length(max(d, 0.0));
}

float sdSphere(vec3 point) {
    return length(point)-1;
}

float sdTriPrism(vec3 point, vec2 h) {
    vec3 q = abs(point);
    return max(q.z-h.y, max(q.x*0.866025+point.y*0.5, -point.y)-h.x*0.5);
}

float calculateDisplacement(vec3 point, float displacement) {
    return sin(displacement * point.x) * sin(displacement * point.y) * sin(displacement * point.z) * 0.25;
}
const float Power = 3.0;
float SDF(vec3 pos, out int steps) {
    vec3 z = pos;
    float dr = 1.0;
    float r = 0.0;
    for (int i = 0; i < 20; i++) {
        r = length(z);
        steps = i;
        if (r>4.0) break;

        // convert to polar coordinates
        float theta = acos(z.z/r);
        float phi = atan(z.y, z.x);
        dr =  pow(r, Power-1.0)*Power*dr + 1.0;

        // scale and rotate the point
        float zr = pow(r, Power);
        theta = theta*Power;
        phi = phi*Power;

        // convert back to cartesian coordinates
        z = zr*vec3(sin(theta)*cos(phi), sin(phi)*sin(theta), cos(theta));
        z+=pos;
    }

    return 0.5*log(r)*r/dr;
}
struct Hit {
    float distance;
    vec3 col;
    float reflectivity;
    float refractivity;
};
//transformace
vec3 translate(vec3 p, vec3 t) {
    return p - t;
}
vec3 scale(vec3 p, vec3 s) {
    mat4 S = mat4(
    vec4(1/s.x, 0, 0, 0),
    vec4(0, 1/s.y, 0, 0),
    vec4(0, 0, 1/s.z, 0),
    vec4(0, 0, 0, 1));
    return (vec4(p, 1)*S).xyz;
}
vec3 rot3D(vec3 p, vec3 axis, float angle) {
    angle = radians(angle);
    return mix(dot(axis, p) * axis, p, cos(angle)) + cross(axis, p) * sin(angle);
}

//definice operátorů pro kombinování objektů
float sdUnion(float o1, float o2) {
    return (o1 < o2) ? o1 : o2;
}
float sdSubstraction(float o1, float o2) {
    o1 = -o1;
    return (o1 > o2) ? o1 : o2;
}
float sdIntersection(float o1, float o2) {
    return (o1 > o2) ? o1 : o2;
}
float sdUnion(float o1, float o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) - k*h*(1.0-h);

}
float sdSubtraction(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2+o1)/k, 0.0, 1.0);
    return mix(o2, -o1, h) + k*h*(1.0-h);
}

float sdIntersection(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) + k*h*(1.0-h);
}
Hit sdUnion(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) - k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);
    float refractivity = mix(o2.refractivity, o1.refractivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;
    obj.refractivity = refractivity;

    return obj;
}
Hit sdSubtraction(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance+o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, -o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}

Hit sdIntersection(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}
//modulo funkce
float pMod (inout float p, float size)
{
    float halfsize = size * 0.5;
    float c = floor((p+halfsize)/size);
    p = mod(p+halfsize, size)-halfsize;
    p = mod(-p+halfsize, size)-halfsize;
    return c;
}
vec3 opTwistY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opTwistX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opTwistZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}

vec3 opBendX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}
vec3 opBendY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opBendZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opSymX(vec3 p)
{
    p.x = abs(p.x);
    return p;
}
vec3 opSymY(vec3 p)
{
    p.y = abs(p.y);
    return p;
}
vec3 opSymZ(vec3 p)
{
    p.z = abs(p.z);
    return p;
}
vec3 opSymXY(vec3 p)
{
    p.xy = abs(p.xy);
    return p;
}
vec3 opSymXZ(vec3 p)
{
    p.xz = abs(p.xz);
    return p;
}
vec3 opSymYZ(vec3 p)
{
    p.yz = abs(p.yz);
    return p;
}
vec3 opSymXYZ(vec3 p)
{
    p = abs(p);
    return p;
}
vec3 repeated(vec3 p, float s)
{
    vec3 r = p - s*round(p/s);
    return r;
}
vec3 refract(vec3 rayDir, vec3 normal, float n1, float n2) {
    float n = n1/n2;
    float cosI = -dot(normal, rayDir);
    float sinT2 = n*n*(1.0-cosI*cosI);
    if (sinT2 > 1.0) {
        return vec3(0.0);
    }
    float cosT = sqrt(1.0-sinT2);
    return n*rayDir + (n*cosI-cosT)*normal;
}
vec3 reflect(vec3 rayDir, vec3 normal) {
    return rayDir - 2.0*dot(rayDir, normal)*normal;
}

////////////////////////////////////////////////////////////////////////


in vec3 vertColor;
out vec4 outColor;

uniform float width;
uniform float height;
//proměná, která se mění s časem
uniform float h;
const int maxNumOfLights = 3;
uniform vec3 camPos;
uniform mat4x4 lookAt;
uniform vec3[maxNumOfLights] lightPos;
uniform vec3[maxNumOfLights] lightCol;
uniform int numOfLights;
uniform sampler2D texture1;
uniform sampler2D texture2;
uniform sampler2D texture3;
uniform int sceneId;


const int steps = 500;
const float minDist = 0.001;
const float maxDist = 100.0;

//maximální počet odrazů paprsku
float maxRec = 5;

layout(origin_upper_left) in vec4 gl_FragCoord;
//výpočet textury
vec3 calculateTexture(sampler2D texture, vec3 point, vec3 normal) {
    vec3 colorXY = texture2D(texture, mod(point.xy*.5+.5, 1)).xyz;
    vec3 colorYZ = texture2D(texture, mod(point.yz*.5+.5, 1)).xyz;
    vec3 colorXZ = texture2D(texture, mod(point.xz*.5+.5, 1)).xyz;
    normal = abs(normal);
    normal = normal*normal*normal;
    return colorXY*normal.z + colorYZ*normal.x +colorXZ*normal.y;
}

float sdTwist(vec3 p) {
    p = opTwistZ(p, 0.5);
    p = scale(p, vec3(1, 1, 4));
    float t = sdBox(p);
    return t;
}
float sdBend(vec3 p) {
    p = opBendX(p, .1);
    p = scale(p, vec3(4, 1, 1));
    float t = sdBox(p);
    return t;
}
float sdDisplacement(vec3 p) {
    float d = calculateDisplacement(p, 5);
    float t = sdBox(p);
    return t+d;
}




Hit mapScene(vec3 point, vec3 normal) {
    float d1 = sdBox(point);
    float d2 = sdSphere(scale(point, vec3(1.2, 1.2, 1.2)));
    Hit box = Hit(d1, vec3(1, 1, 1), 0, 0);
    Hit sphere = Hit(d2, vec3(1, 0, 0), 0, 0);
    Hit b = sdSubtraction(sphere, box, 0);
    vec3 p = translate(point, vec3(0, 16, 0));
    p = scale(p, vec3(2, 2, 2));
    float sphere1 = sdSphere(p);
    Hit sphereHit = Hit(sphere1, vec3(0.5, 0, 0), 1, 0);
    vec3 q = translate(point, vec3(0, 0, 4));

    float plane = sdPlane(q, vec4(0, 0, -1, 0));
    Hit planeHit = Hit(plane, calculateTexture(texture1, q, normal), 0, 0);


    float box1 = sdBox(translate(point, vec3(16, 16, 0)));
    Hit boxHit = Hit(box1, vec3(1, 1, 1), 0, 1);
    Hit scene = sdUnion(sphereHit, boxHit, 0);

    vec3 t = translate(point, vec3(-16, 16, 0));
    float twist = sdTwist(t);
    Hit twistHit = Hit(twist, vec3(1, 1, 1), 0, 0);
    scene = sdUnion(scene, twistHit, 0);
    vec3 f = translate(point, vec3(-16, 0, 0));
    float bend = sdBend(f);
    Hit bendHit = Hit(bend, vec3(1, .5, .5), 0, 0);
    scene = sdUnion(scene, bendHit, 0);

    vec3 s = translate(point, vec3(16, 0, 0));
    float blob = sdSphere(translate(s, vec3(-1, .5, 0)));
    Hit blobHit = Hit(blob, vec3(1, 0, 0), 0, 0);
    blobHit = sdUnion(blobHit, Hit(sdSphere(translate(s, vec3(0, 1, 0))), vec3(0, 1, 0), 0, 0), 0.5);
    blobHit = sdUnion(blobHit, Hit(sdSphere(translate(s, vec3(.5, 0, -.5))), vec3(0, 0, 1), 0, 0), 0.5);
    scene = sdUnion(scene, blobHit, 0);

    float displacement = sdDisplacement(translate(point, vec3(0, -16, 0)));
    Hit displacementHit = Hit(displacement, vec3(1, 1, 1), 0, 0);
    scene = sdUnion(scene, displacementHit, 0);

    int i = 0;
    float frac = SDF(translate(point, vec3(-16, -16, 0)), i);
    Hit fracHit = Hit(frac, vec3(i/8, 1, 1-i/8), 0, 0);
    scene = sdUnion(scene, fracHit, 0);

    float octahedron = sdOctahedron(translate(point, vec3(16, -16, 0)), 1);
    Hit octahedronHit = Hit(octahedron, vec3(0, 1, 1), 0, 1);
    scene = sdUnion(scene, octahedronHit, 0);

    scene = sdUnion(scene, b, 0);
    return sdUnion(scene, planeHit, 0);

}
float preMapScene(vec3 point) {
    float d1 = sdBox(point);
    float d2 = sdSphere(scale(point, vec3(1.2, 1.2, 1.2)));
    float b = sdSubtraction(d2, d1, 0);
    vec3 p = translate(point, vec3(0, 16, 0));
    p = scale(p, vec3(2, 2, 2));
    float sphere = sdSphere(p);
    vec3 q = translate(point, vec3(0, 0, 4));

    float plane = sdPlane(q, vec4(0, 0, -1, 0));

    float box = sdBox(translate(point, vec3(16, 16, 0)));
    float scene = sdUnion(sphere, box, 0);
    vec3 t = translate(point, vec3(-16, 16, 0));
    float twist = sdTwist(t);
    scene = sdUnion(scene, twist, 0);
    vec3 f = translate(point, vec3(-16, 0, 0));
    float bend = sdBend(f);
    scene = sdUnion(scene, bend, 0);
    vec3 s = translate(point, vec3(16, 0, 0));
    float blob = sdSphere(translate(s, vec3(-1, .5, 0)));
    blob = sdUnion(blob, sdSphere(translate(s, vec3(0, 1, 0))), 0.5);
    blob = sdUnion(blob, sdSphere(translate(s, vec3(0.5, 0, -.5))), 0.5);
    scene = sdUnion(scene, blob, 0);

    float displacement = sdDisplacement(translate(point, vec3(0, -16, 0)));
    scene = sdUnion(scene, displacement, 0);

    int i = 0;
    float frac = SDF(translate(point, vec3(-16, -16, 0)), i);
    scene = sdUnion(scene, frac, 0);

    float octahedron = sdOctahedron(translate(point, vec3(16, -16, 0)), 1);
    scene = sdUnion(scene, octahedron, 0);

    scene = sdUnion(scene, b, 0);
    return sdUnion(scene, plane, 0);

}
float preMapSceneInverted(vec3 point) {
    vec3 p = translate(point, vec3(16, 16, 0));
    float octahedron = sdOctahedron(translate(point, vec3(16, -16, 0)), 1)*-1;
    float box = sdBox(p)*-1;
    return sdIntersection(box, octahedron, 0);
}

//výpočet normály
vec3 calculateNormal(vec3 point)
{
    const vec2 epsylon = vec2(1e-3, 0);
    vec3 normal = vec3(
    preMapScene(point+epsylon.xyy) - preMapScene(point-epsylon.xyy),
    preMapScene(point+epsylon.yxy) - preMapScene(point-epsylon.yxy),
    preMapScene(point+epsylon.yyx) - preMapScene(point-epsylon.yyx))/epsylon.x;

    return normalize(normal);
}

float hardShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt)
{
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        t += h;
    }
    return 1.0;
}

float softShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt, float k)
{
    float result = 1.0;
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        result = min(result, k*h/t);
        t += h;
    }
    return result;
}

float ambientOcclusion(vec3 p, vec3 n)
{
    float aoStepSize = 0.2;
    float aoIterations = 3;
    float aoIntensity = 0.25;
    float step = aoStepSize;
    float ao = 0.0;
    float dist;
    for (int i = 1; i <= aoIterations; i++)
    {
        dist = step * i;
        ao += max(0.0, (dist - preMapScene(p + n * dist)) / dist);
    }
    return (1.0 - ao * aoIntensity);
}

Hit shading(vec3 point, vec3 normal, Hit obj, vec3 cameraDir)
{
    vec3 color = obj.col;
    vec3 result;
    float lightIntensity = 2;
    float shadowIntensity = 1;
    vec3 light;
    vec3 lightFinal = vec3(0, 0, 0);
    float shadow;
    vec3 lightColor = vec3(0, 0, 0);
    vec3 dirToLight;
    for (int i = 0; i < numOfLights; i++) {
        dirToLight = normalize(lightPos[i] - point);
        float diffuseStrength = max(0.0, dot(dirToLight, normal));
        vec3 halfway = normalize(dirToLight - cameraDir);
        float specularStrength = max(0.0, dot(normal, halfway));
        specularStrength = pow(specularStrength, 32);

        light = lightCol[i] * (diffuseStrength*0.333+specularStrength*0.333) * lightIntensity;
        //Vržené stíny
        //        shadow = hardShadow(point, dirToLight, 0.1, 8) * 0.5 + 0.5;
        shadow = softShadow(point, dirToLight, 0.1, 12, 8) * 0.5 + 0.5;
        shadow = max(0.0, pow(shadow, shadowIntensity));
        light *= shadow;
        lightColor += lightCol[i];
        lightFinal += light;
    }
    lightColor /= numOfLights;
    float ambientStrenght = 0.5;
    vec3 lightAmbient = ambientStrenght*0.333*lightColor;
    float ambientOcc = ambientOcclusion(point, normal);
    result = color * ambientOcc;

    result *= lightFinal + lightAmbient;
    obj.col = result;
    return obj;
}


Hit rayMarching(vec3 rayOri, vec3 rayDir, inout vec3 currentPos, out vec3 normal) {
    float distTraveled = 0.0;
    const float overRelaxation = 1.5;

    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }

        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně

        float distToClosest = preMapScene(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {
            normal = calculateNormal(currentPos);
            Hit hit = mapScene(currentPos, normal);
            //            return hit;
            return shading(currentPos, normal, hit, rayDir);
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return Hit(0.0, vec3(0.5), 0.0, 0.0);
}
float rayMarchingInverted(vec3 rayOri, vec3 rayDir, inout vec3 currentPos, out vec3 normal) {
    float distTraveled = 0.0;


    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }

        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně

        float distToClosest = preMapSceneInverted(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {
            normal = calculateNormal(currentPos)*-1;

            //                        return shading(currentPos, normal, getMaterial(currentPos, Hit.y, normal, reflective));
            //                        return Hit.col;
            return distTraveled;
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return distTraveled;
}


vec3 getResult(vec3 rayOri, vec3 rayDir) {
    vec3 normal;
    vec3 currentPos;
    Hit hit = rayMarching(rayOri, rayDir, currentPos, normal);
    vec3 color = hit.col;
    float reflective = hit.reflectivity;
    float refractive = hit.refractivity;
    int iterations = 0;
    //pokud je povrch zrdcadlový
    for (int i = 0; i < maxRec; i++) {
        iterations++;
        if (refractive > 0) {
            vec3 n = normal;
            vec3 pos = currentPos;
            vec3 dir = rayDir;
            vec3 refractedIn = refract(rayDir, normal, 1.0, 1.5);
            vec3 rayOriIn = currentPos - normal*0.02;
            float dIn = rayMarchingInverted(rayOriIn, refractedIn, currentPos, normal);
            vec3 refractedOut = refract(refractedIn, normal, 1.5, 1.0);
            if (dot(refractedOut, refractedOut) < 0.0000001) {
                refractedOut = reflect(refractedIn, normal);
                vec3 rayOriOut = rayOri - normal*0.02;
                Hit dOut = rayMarching(rayOriOut, refractedOut, currentPos, normal);
                float density = 0.5;
                float opticalDist = exp(-dIn*density);

                color += dOut.col*opticalDist*refractive;
                refractive = dOut.refractivity;
                reflective = dOut.reflectivity;
            } else {
                vec3 rayOriOut = currentPos - normal*0.02;
                Hit dOut = rayMarching(rayOriOut, refractedOut, currentPos, normal);
                float density = 0.5;
                float opticalDist = exp(-dIn*density);
                color += dOut.col*opticalDist*refractive;
                refractive = dOut.refractivity;
                reflective = dOut.reflectivity;
            }
            float fresnel = pow(1.0+dot(rayDir, n), 5.0);
            vec3 reflected = reflect(dir, n);
            Hit o = rayMarching(pos, reflected, pos, n);
            color = mix(color, o.col, fresnel);
            rayDir = refractedOut;
        }
        else if (reflective > 0) {
            currentPos += normal*0.02;
            vec3 reflected = reflect(rayDir, normal);
            Hit o = rayMarching(currentPos, reflected, currentPos, normal);
            color += o.col*reflective;
            reflective = o.reflectivity;
            refractive = o.refractivity;
            rayDir = reflected;
        }
        else {
            break;
        }

    }
    return color/iterations;
}


void main() {
    //řešení velikosti okna a poměru stran
    vec2 uv = vec2(((gl_FragCoord.x - 0.5) / width)*2-1, ((gl_FragCoord.y - 0.5) / height)*2-1);
    float aspectRatio = width/height;
    if (aspectRatio < 1) {
        uv.y *= 1/aspectRatio;
    } else {
        uv.x *= aspectRatio;
    }
    //počáteční pozice paprsku
    vec3 rayOri = camPos;
    //směr paprsku
    vec4 temp =vec4(uv, 1, 1)*lookAt;
    vec3 rayDir = normalize(temp.xyz);
    //výsledná barva pixelu
    vec3 color = getResult(rayOri, rayDir);
    outColor = vec4(color, 1.0);
}

