#version 330
//definice funkcí pro výpočet vzdálenosti od objektu

float sdPlane(vec3 point, vec4 n) {
    //n musí být normalizované
    return dot(point, n.xyz) + n.w;
}
float sdTorus(vec3 p, vec2 t)
{
    vec2 q = vec2(length(p.xy)-t.x, p.z);
    return length(q)-t.y;
}
float sdVerticalCapsule(vec3 p, float h, float r)
{
    p.y -= clamp(p.y, 0.0, h);
    return length(p) - r;
}

float sdBox(vec3 point) {
    vec3 d = abs(point)-1;
    return min(max(d.x, max(d.y, d.z)), 0.0) + length(max(d, 0.0));
}

float sdRoundBox(vec3 p, float r)
{
    vec3 q = abs(p)-1;
    return length(max(q, 0.0)) + min(max(q.x, max(q.y, q.z)), 0.0) - r;
}

float sdSphere(vec3 point) {
    return length(point)-1;
}

float sdTriPrism(vec3 point, vec2 h) {
    vec3 q = abs(point);
    return max(q.z-h.y, max(q.x*0.866025+point.y*0.5, -point.y)-h.x*0.5);
}

float calculateDisplacement(vec3 point, float displacement) {
    return sin(displacement * point.x) * sin(displacement * point.y) * sin(displacement * point.z) * 0.25;
}

struct Hit {
    float distance;
    vec3 col;
    float reflectivity;
    float refractivity;
};
//transformace
vec3 translate(vec3 p, vec3 t) {
    return p - t;
}
vec3 scale(vec3 p, vec3 s) {
    mat4 S = mat4(
    vec4(1/s.x, 0, 0, 0),
    vec4(0, 1/s.y, 0, 0),
    vec4(0, 0, 1/s.z, 0),
    vec4(0, 0, 0, 1));
    return (vec4(p, 1)*S).xyz;
}
vec3 rot3D(vec3 p, vec3 axis, float angle) {
    angle = radians(angle);
    return mix(dot(axis, p) * axis, p, cos(angle)) + cross(axis, p) * sin(angle);
}

//definice operátorů pro kombinování objektů
float sdUnion(float o1, float o2) {
    return (o1 < o2) ? o1 : o2;
}
float sdSubstraction(float o1, float o2) {
    o1 = -o1;
    return (o1 > o2) ? o1 : o2;
}
float sdIntersection(float o1, float o2) {
    return (o1 > o2) ? o1 : o2;
}
float sdUnion(float o1, float o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) - k*h*(1.0-h);

}
float sdSubtraction(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2+o1)/k, 0.0, 1.0);
    return mix(o2, -o1, h) + k*h*(1.0-h);
}

float sdIntersection(float o1, float o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2-o1)/k, 0.0, 1.0);
    return mix(o2, o1, h) + k*h*(1.0-h);
}
Hit sdUnion(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 + 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) - k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);
    float refractivity = mix(o2.refractivity, o1.refractivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;
    obj.refractivity = refractivity;

    return obj;
}
Hit sdSubtraction(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance+o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, -o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}

Hit sdIntersection(Hit o1, Hit o2, float k)
{
    float h = clamp(0.5 - 0.5*(o2.distance-o1.distance)/k, 0.0, 1.0);
    float dist = mix(o2.distance, o1.distance, h) + k*h*(1.0-h);
    vec3 col = mix(o2.col, o1.col, h);
    float reflectivity = mix(o2.reflectivity, o1.reflectivity, h);

    Hit obj;
    obj.distance = dist;
    obj.col = col;
    obj.reflectivity = reflectivity;

    return obj;
}
//modulo funkce
float pMod (inout float p, float size)
{
    float halfsize = size * 0.5;
    float c = floor((p+halfsize)/size);
    p = mod(p+halfsize, size)-halfsize;
    p = mod(-p+halfsize, size)-halfsize;
    return c;
}
vec3 opTwistY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opTwistX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opTwistZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}

vec3 opBendX(vec3 p, float k)
{
    float c = cos(k*p.x);
    float s = sin(k*p.x);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xy, p.z);
    return q;
}
vec3 opBendY(vec3 p, float k)
{
    float c = cos(k*p.y);
    float s = sin(k*p.y);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(m*p.xz, p.y);
    return q;
}
vec3 opBendZ(vec3 p, float k)
{
    float c = cos(k*p.z);
    float s = sin(k*p.z);
    mat2  m = mat2(c, -s, s, c);
    vec3  q = vec3(p.x, m*p.yz);
    return q;
}
vec3 opSymX(vec3 p)
{
    p.x = abs(p.x);
    return p;
}
vec3 opSymY(vec3 p)
{
    p.y = abs(p.y);
    return p;
}
vec3 opSymZ(vec3 p)
{
    p.z = abs(p.z);
    return p;
}
vec3 opSymXY(vec3 p)
{
    p.xy = abs(p.xy);
    return p;
}
vec3 opSymXZ(vec3 p)
{
    p.xz = abs(p.xz);
    return p;
}
vec3 opSymYZ(vec3 p)
{
    p.yz = abs(p.yz);
    return p;
}
vec3 opSymXYZ(vec3 p)
{
    p = abs(p);
    return p;
}
vec3 repeated(vec3 p, float s)
{
    vec3 r = p - s*round(p/s);
    return r;
}
//vec3 opRepX( vec3 p, float k )
//{
//    return mod(p,k)-k*0.5;
//}
vec3 refract(vec3 rayDir, vec3 normal, float n1, float n2) {
    float n = n1/n2;
    float cosI = -dot(normal, rayDir);
    float sinT2 = n*n*(1.0-cosI*cosI);
    if (sinT2 > 1.0) {
        return vec3(0.0);
    }
    float cosT = sqrt(1.0-sinT2);
    return n*rayDir + (n*cosI-cosT)*normal;
}
vec3 reflect(vec3 rayDir, vec3 normal) {
    return rayDir - 2.0*dot(rayDir, normal)*normal;
}

////////////////////////////////////////////////////////////////////////


in vec3 vertColor;
out vec4 outColor;

uniform float width;
uniform float height;
//proměná, která se mění s časem
uniform float h;
const int maxNumOfLights = 3;
uniform vec3 camPos;
uniform mat4x4 lookAt;
uniform vec3[maxNumOfLights] lightPos;
uniform vec3[maxNumOfLights] lightCol;
uniform int numOfLights;
uniform sampler2D texture1;
uniform sampler2D texture2;
uniform sampler2D texture3;
uniform int sceneId;


const int steps = 500;
const float minDist = 0.001;
const float maxDist = 300.0;

//maximální počet odrazů paprsku
float maxRec = 5;

layout(origin_upper_left) in vec4 gl_FragCoord;
//výpočet textury
vec3 calculateTexture(sampler2D texture, vec3 point, vec3 normal) {
    vec3 colorXY = texture2D(texture, mod(point.xy*.5+.5, 1)).xyz;
    vec3 colorYZ = texture2D(texture, mod(point.yz*.5+.5, 1)).xyz;
    vec3 colorXZ = texture2D(texture, mod(point.xz*.5+.5, 1)).xyz;
    normal = abs(normal);
    normal = normal*normal*normal;
    return colorXY*normal.z + colorYZ*normal.x +colorXZ*normal.y;
}

float sdLence(vec3 point) {
    vec3 p = opSymX(point);
    p = translate(p, vec3(15, 0, 0));
    p = scale(p, vec3(12, 1, 8));
    float s1 = sdSphere(p);
    return s1;
}
float sdHoops(vec3 point) {
    //make hoops for glasses
    vec3 p = opSymX(point);
    vec3 q = translate(p, vec3(1, 0, -5));
    q = scale(q, vec3(1, 0.2, .7));
    float bridge = sdRoundBox(q, 2);
    p = translate(p, vec3(15, 0, 0));
    p = scale(p, vec3(4.5, 0.15, 3));
    float s1 = sdRoundBox(p, 2);
    vec3 e = translate(p, vec3(15, 0, 5));
    e = scale(e, vec3(40, 10, 3));
    float clippingBox = sdBox(e);
    s1 = sdSubtraction(clippingBox, s1, 2);
    s1 = sdUnion(s1, bridge, 2);
    return s1;
}
float sdLegs(vec3 point) {
    vec3 p = opSymX(point);
    p = translate(p, vec3(25, 29.5, -6));
    vec3 q = translate(p, vec3(0, 36.5, 6));
    q = opBendZ(q, .05);
    q = rot3D(q, vec3(1, 0, 0), -45);
    q = scale(q, vec3(.7, 4, 1));

    float s2 = sdRoundBox(q, 1);
    p = scale(p, vec3(.7, 15, 1));
    float s1 = sdRoundBox(p, 1);
    s1 = sdUnion(s1, s2, 3);
    return s1;
}
float sdMirror(vec3 point) {
    point = translate(point, vec3(-75, -40, 0));
    point = rot3D(point, vec3(0, 0, 1), -45);
    point = scale (point, vec3(2, 30, 30));
    float box = sdBox(point);
    return box;
}
float sdPhone(vec3 point) {
    float box = sdRoundBox(point,1);
    return box;
}
float sdScreen(vec3 point) {
    vec3 q = scale(point, vec3(1.5, 1.5, .5));
    q = translate(q, vec3(0, 0, -4));
    float box = sdBox(q);
    return box;
}

Hit mapScene(vec3 point, vec3 normal) {
    float lences = sdLence(point);
    Hit lencesHit = Hit(lences, vec3(1, 1, 1), 0, 1);
    float hoops = sdHoops(point);
    Hit hoopsHit = Hit(hoops, vec3(0, 0, 0), 0, 0);
    float legs = sdLegs(point);
    Hit legsHit = Hit(legs, vec3(0, 0, 0), 0, 0);
    Hit scene = sdUnion(legsHit, hoopsHit, 4);
    scene = sdUnion(lencesHit, scene, 0);



    vec3 q = translate(point, vec3(0, 0, 8));
    float plane = sdPlane(q, vec4(0, 0, -1, 0));
    Hit planeHit = Hit(plane, calculateTexture(texture3, scale(q, vec3(100, 100, 0)), normal), 0, 0);
    q = rot3D(q, vec3(0, 0, 1), 45);
    q = scale(q, vec3(.5, .5, 1));
    vec3 p = translate(q, vec3(50, -75, 0));
    p = scale(p, vec3(30, 15, 3));
    float phone = sdPhone(p);
    Hit phoneHit = Hit(phone, vec3(.1, .1, .1), 0, 0);
    float screen = sdScreen(p);
    vec3 t = translate(q, vec3(50, 25, 0));
    t = rot3D(t, vec3(0, 0, 1), -90);
    t = scale(t, vec3(25, 45, 0));
    Hit screenHit = Hit(screen, calculateTexture(texture2, t, normal), 0, 0);
    Hit temp = sdUnion(phoneHit, screenHit, 0);
    scene = sdUnion(scene, temp, 0);

    Hit boxHit = Hit(sdMirror(point), vec3(.5, .5, .5), 1, 0);

    scene = sdUnion(scene, boxHit, 0);

    return sdUnion(scene, planeHit, 0);

}
float preMapScene(vec3 point) {
    float lences = sdLence(point);
    float hoops = sdHoops(point);
    float legs = sdLegs(point);
    float scene = sdUnion(legs, hoops, 4);
    scene = sdUnion(lences, scene, 0);


    vec3 q = translate(point, vec3(0, 0, 8));
    float plane = sdPlane(q, vec4(0, 0, -1, 0));
    q = rot3D(q, vec3(0, 0, 1), 45);
    q = scale(q, vec3(.5, .5, 1));
    vec3 p = translate(q, vec3(50, -75, 0));
    p = scale(p, vec3(30, 15, 3));
    float phone = sdPhone(p);
    scene = sdUnion(scene, phone, 0);

    float box = sdMirror(point);
    scene = sdUnion(scene, box, 0);
    return sdUnion(scene, plane, 0);

}
float preMapSceneInverted(vec3 point) {
    float sphere = sdLence(point)*-1;
    return sphere;
}

//výpočet normály
vec3 calculateNormal(vec3 point)
{
    const vec2 epsylon = vec2(1e-3, 0);
    vec3 normal = vec3(
    preMapScene(point+epsylon.xyy) - preMapScene(point-epsylon.xyy),
    preMapScene(point+epsylon.yxy) - preMapScene(point-epsylon.yxy),
    preMapScene(point+epsylon.yyx) - preMapScene(point-epsylon.yyx))/epsylon.x;

    return normalize(normal);
}

float hardShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt)
{
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        t += h;
    }
    return 1.0;
}

float softShadow(vec3 rayOri, vec3 rayDir, float mint, float maxt, float k)
{
    float result = 1.0;
    for (float t = mint; t < maxt;)
    {
        float h = preMapScene(rayOri+rayDir*t);
        if (h < 0.001)
        {
            return 0.0;
        }
        result = min(result, k*h/t);
        t += h;
    }
    return result;
}

float ambientOcclusion(vec3 p, vec3 n)
{
    float aoStepSize = 0.2;
    float aoIterations = 3;
    float aoIntensity = 0.25;
    float step = aoStepSize;
    float ao = 0.0;
    float dist;
    for (int i = 1; i <= aoIterations; i++)
    {
        dist = step * i;
        ao += max(0.0, (dist - preMapScene(p + n * dist)) / dist);
    }
    return (1.0 - ao * aoIntensity);
}

Hit shading(vec3 point, vec3 normal, Hit obj, vec3 cameraDir)
{
    vec3 color = obj.col;
    vec3 result;
    float lightIntensity = 5;
    float shadowIntensity = 1;
    vec3 light;
    vec3 lightFinal = vec3(0, 0, 0);
    float shadow;
    vec3 lightColor = vec3(0, 0, 0);
    vec3 dirToLight;
    for (int i = 0; i < numOfLights; i++) {
        dirToLight = normalize(lightPos[i] - point);
        float diffuseStrength = max(0.0, dot(dirToLight, normal));
        vec3 halfway = normalize(dirToLight - cameraDir);
        float specularStrength = max(0.0, dot(normal, halfway));
        specularStrength = pow(specularStrength, 32);

        light = lightCol[i] * (diffuseStrength*0.333+specularStrength*0.333) * lightIntensity;
        //Vržené stíny
//        shadow = hardShadow(point, dirToLight, 0.1, 8) * 0.5 + 0.5;
                shadow = softShadow(point, dirToLight, 0.1, 12, 8) * 0.5 + 0.5;
        shadow = max(0.0, pow(shadow, shadowIntensity));
        light *= shadow;
        lightColor += lightCol[i];
        lightFinal += light;
    }
    lightColor /= numOfLights;
    float ambientStrenght = 0.5;
    vec3 lightAmbient = ambientStrenght*0.333*lightColor;
    float ambientOcc = ambientOcclusion(point, normal);
    result = color * ambientOcc;

    result *= lightFinal + lightAmbient;
    obj.col = result;
    return obj;
}


Hit rayMarching(vec3 rayOri, vec3 rayDir, inout vec3 currentPos, out vec3 normal) {
    float distTraveled = 0.0;
    const float overRelaxation = 1.5;

    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }

        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně

        float distToClosest = preMapScene(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {
            normal = calculateNormal(currentPos);
            Hit hit = mapScene(currentPos, normal);
            //            return hit;
            return shading(currentPos, normal, hit, rayDir);
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return Hit(0.0, vec3(0.5), 0.0, 0.0);
//    return Hit(0.0, calculateTexture(texture1, scale(currentPos, vec3(100, 100, 0)), -rayDir), 0.0, 0.0);
}
float rayMarchingInverted(vec3 rayOri, vec3 rayDir, inout vec3 currentPos, out vec3 normal) {
    float distTraveled = 0.0;


    //Ray Marching cyklus
    for (int i = 0; i < steps; i++)
    {
        //pokud je překonána maximální vzdálenost, nemá smysl pokračovat
        if (distTraveled > maxDist)
        {
            break;
        }

        currentPos = rayOri + rayDir * distTraveled;
        //nejbližší objekt ve scéně

        float distToClosest = preMapSceneInverted(currentPos);
        //pokud je objekt blíže než je minDist, je pixel vybarven
        if (distToClosest < minDist)
        {
            normal = calculateNormal(currentPos)*-1;

            //                        return shading(currentPos, normal, getMaterial(currentPos, Hit.y, normal, reflective));
            //                        return Hit.col;
            return distTraveled;
        }
        distTraveled += distToClosest;
    }
    //pokud nebylo nic zasaženo, barva pozadí
    return distTraveled;
}


vec3 getResult(vec3 rayOri, vec3 rayDir) {
    vec3 normal;
    vec3 currentPos;
    Hit hit = rayMarching(rayOri, rayDir, currentPos, normal);
    vec3 color = hit.col;
    float reflective = hit.reflectivity;
    float refractive = hit.refractivity;
    int iterations = 0;
    //pokud je povrch zrdcadlový
    for (int i = 0; i < maxRec; i++) {
        iterations++;
        if (refractive > 0) {
            vec3 n = normal;
            vec3 pos = currentPos;
            vec3 dir = rayDir;
            vec3 refractedIn = refract(rayDir, normal, 1.0, 1.52);
            vec3 rayOriIn = currentPos - normal*0.02;
            float dIn = rayMarchingInverted(rayOriIn, refractedIn, currentPos, normal);
            vec3 refractedOut = refract(refractedIn, normal, 1.52, 1.0);
            if (dot(refractedOut, refractedOut) < 0.0000001) {
                refractedOut = reflect(refractedIn, normal);
                vec3 rayOriOut = rayOri - normal*0.02;
                Hit dOut = rayMarching(rayOriOut, refractedOut, currentPos, normal);
                float density = 0.05;
                float opticalDist = exp(-dIn*density);

                color += dOut.col*opticalDist*refractive;
                refractive = dOut.refractivity;
                reflective = dOut.reflectivity;
            } else {
                vec3 rayOriOut = currentPos - normal*0.02;
                Hit dOut = rayMarching(rayOriOut, refractedOut, currentPos, normal);
                float density = 0.05;
                float opticalDist = exp(-dIn*density);
                color += dOut.col*opticalDist*refractive;
                refractive = dOut.refractivity;
                reflective = dOut.reflectivity;
            }
            float fresnel = pow(1.0+dot(rayDir, n), 5.0);
            vec3 reflected = reflect(dir, n);
            Hit o = rayMarching(pos, reflected, pos, n);
            color = mix(color, o.col, fresnel);
            rayDir = refractedOut;
        }
        else if (reflective > 0) {
            currentPos += normal*0.02;
            vec3 reflected = reflect(rayDir, normal);
            Hit o = rayMarching(currentPos, reflected, currentPos, normal);
            color += o.col*reflective;
            reflective = o.reflectivity;
            refractive = o.refractivity;
            rayDir = reflected;
        }
        else {
            break;
        }

    }
    return color/iterations;
}


void main() {
    //řešení velikosti okna a poměru stran
    vec2 uv = vec2(((gl_FragCoord.x - 0.5) / width)*2-1, ((gl_FragCoord.y - 0.5) / height)*2-1);
    float aspectRatio = width/height;
    if (aspectRatio < 1) {
        uv.y *= 1/aspectRatio;
    } else {
        uv.x *= aspectRatio;
    }
    //počáteční pozice paprsku
    vec3 rayOri = camPos;
    //směr paprsku
    vec4 temp =vec4(uv, 1, 1)*lookAt;
    vec3 rayDir = normalize(temp.xyz);
    //výsledná barva pixelu
    vec3 color = getResult(rayOri, rayDir);
    outColor = vec4(color, 1.0);
}

