package model;

import lwjglutils.OGLTextRenderer;
import ui.components.UIButton;
import ui.components.UIComponent;
import ui.components.UIInputField;
import ui.utils.UITextProvider;

import java.util.ArrayList;
import java.util.List;

import static org.lwjgl.glfw.GLFW.*;


public class Canvas {
    private OGLTextRenderer textRenderer;
    private List<UIComponent> components;
    private List<UIButton> buttons;
    private List<UIInputField> inputFields;
    private UIComponent focusedComponent;
    int width, height;
    public Canvas(int width, int height, OGLTextRenderer textRenderer) {
        this.textRenderer = textRenderer;
        components = new ArrayList<>();
        buttons = new ArrayList<>();
        inputFields = new ArrayList<>();
        this.width = width;
        this.height = height;
    }
    public void render() {
        for(UIComponent component : components)
            component.render();
        for (UIComponent component : components)
            if (component instanceof UITextProvider)
                ((UITextProvider) component).renderText();
    }
    public void resize(int width, int height) {
        this.width = width;
        this.height = height;
        for (UIComponent component : components)
            component.resize(width, height);
        textRenderer.resize(width, height);

    }
    public void checkButtons(int button, int action,float posX,float posY) {
        boolean focused = false;
        if (action == GLFW_PRESS && button == GLFW_MOUSE_BUTTON_1) {
            for (UIButton uiButton : buttons)
                uiButton.checkButton(button, action, posX, posY);
            for (UIInputField uiInputField : inputFields)
                if(uiInputField.checkField(button, action, posX, posY)) {
                    focusedComponent = uiInputField;
                    focused = true;
                    break;
                }
            if (!focused)
                focusedComponent = null;
        }
        if (action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_1) {
            for (UIButton uiButton : buttons)
                uiButton.release();
        }

    }
    public void addComponent(UIComponent component) {
        components.add(component);
        if (component instanceof UIButton)
            buttons.add((UIButton) component);
        else if (component instanceof UIInputField)
            inputFields.add((UIInputField) component);
    }

    public OGLTextRenderer getTextRenderer() {
        return textRenderer;
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }


    public void keyTyped(int codepoint) {
        if (focusedComponent != null && focusedComponent instanceof UIInputField)
            ((UIInputField)focusedComponent).keyTyped(codepoint);
    }

    public void keyPressed(int key) {
        if (focusedComponent != null && focusedComponent instanceof UIInputField)
            ((UIInputField)focusedComponent).keyPressed(key);
    }
}
