package model;

import transforms.*;

import java.util.ArrayList;
import java.util.List;

public class Scene {
    private Camera camera;
    private List<Point3D> lightPos = new ArrayList<>();
    private List<Vec3D> lightCol = new ArrayList<>();
    private double speed = 0.1;
    private float time = 0;

    public Scene() {
        init();
    }

    public void init() {
        camera = new Camera(new Vec3D(5,0,0), 0, 0, 1, true);

        lightPos.add(new Point3D(0, 10, -40));
        lightCol.add(new Vec3D(1,1,1));

    }


    public Camera getCamera() {
        return camera;
    }

    public void setCamera(Camera camera) {
        this.camera = camera;
    }

    public void move(boolean[] b) {
        if (b[6]) {
            speed = .5;
        } else {
            speed = 0.1;
        }
        if (b[0]) {
            camera = camera.forward(-speed);
        }
        if (b[1]) {
            camera = camera.backward(-speed);
        }
        if (b[2]) {
            camera = camera.left(speed);
        }
        if (b[3]) {
            camera = camera.right(speed);
        }
        if (b[4]) {
            camera = camera.up(-speed);
        }
        if (b[5]) {
            camera = camera.down(-speed);
        }
    }
    public void rotate(Vec2D change) {
        double azimuth = Math.toRadians(change.getX()*0.1);
        double zenith = Math.toRadians(change.getY()*0.1);
        camera = camera.addAzimuth(azimuth);
        camera = camera.addZenith(-zenith);


    }

    public float[] getLightPos() {
        float[] light =  new float[lightPos.size()*3];
        for (int i = 0; i < lightPos.size(); i++) {
            int k = i * 3;
            light[k] = (float)lightPos.get(i).getX();
            light[k+1] = (float)lightPos.get(i).getY();
            light[k+2] = (float)lightPos.get(i).getZ();
        }
        return light;
    }
    public float[] getLightCol() {
        float[] light =  new float[lightCol.size()*3];
        for (int i = 0; i < lightCol.size(); i++) {
            int k = i * 3;
            light[k] = (float)lightCol.get(i).getX();
            light[k+1] = (float)lightCol.get(i).getY();
            light[k+2] = (float)lightCol.get(i).getZ();
        }
        return light;
    }

    public int getNumOfLights() {
        return lightPos.size();
    }

    public void rotateAround(float deltaTime) {
        time += deltaTime;
        Vec3D a = new Vec3D(Math.sin(time)*10,-Math.cos(time)*10,0);
        camera = camera.withPosition(a);

        camera = camera.withAzimuth(time+29.85);
    }

    public void setMultipleLights(boolean b) {
        if (b) {
            lightPos.add(new Point3D(-10, 0, -40));
            lightCol.add(new Vec3D(.33,.66,0));
        } else {
            if(lightPos.size() > 1) {
                lightPos.remove(1);
                lightCol.remove(1);
            }
        }
    }
    public void changeColor() {
        Vec3D currentColor = lightCol.get(1);
        double red = currentColor.getX();
        double green = currentColor.getY();
        double blue = currentColor.getZ();

        red = (red + .001) % 1;
        green = (green + .001) % 1;
        blue = (blue + .001) % 1;

        // Nastavení nové barvy
        currentColor = new Vec3D(red, green, blue);
        lightCol.set(1, currentColor);
    }
}
