package model;

import renderer.AbstractRenderer;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.MemoryStack;

import java.nio.IntBuffer;

import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.system.MemoryStack.stackPush;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Window {

    public static int WIDTH = 600;
    public static int HEIGHT = 400;

    public static long window;
    private AbstractRenderer renderer;


    public Window(AbstractRenderer renderer) {
        this(WIDTH, HEIGHT, renderer);
    }

    public Window(int width, int height, AbstractRenderer renderer) {
        this.renderer = renderer;
        WIDTH = width;
        HEIGHT = height;
        run();
    }

    public void run() {
        init();

        loop();

        renderer.dispose();
        glfwFreeCallbacks(window);
        glfwDestroyWindow(window);
        glfwTerminate();
        glfwSetErrorCallback(null).free();
    }

    private void init() {
        GLFWErrorCallback.createPrint(System.err).set();
        if (!glfwInit())
            throw new IllegalStateException("Nepodařilo se inicializovat GLFW");
        glfwDefaultWindowHints();
        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE);
        glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE);
        String text = "Ray Marching - Daniel Paleček";
        window = glfwCreateWindow(WIDTH, HEIGHT, text, NULL, NULL);
        if (window == NULL)
            throw new RuntimeException("Selhalo vytvoření GLFW okna");

        glfwSetKeyCallback(window, renderer.getGlfwKeyCallback());
        glfwSetWindowSizeCallback(window, renderer.getGlfwWindowSizeCallback());
        glfwSetMouseButtonCallback(window, renderer.getGlfwMouseButtonCallback());
        glfwSetCursorPosCallback(window, renderer.getGlfwCursorPosCallback());
        glfwSetScrollCallback(window, renderer.getGlfwScrollCallback());
        glfwSetCharCallback(window, renderer.getGlfwCharCallback());

        try (MemoryStack stack = stackPush()) {
            IntBuffer pWidth = stack.mallocInt(1);
            IntBuffer pHeight = stack.mallocInt(1);
            glfwGetWindowSize(window, pWidth, pHeight);

            GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());

            glfwSetWindowPos(
                    window,
                    (vidmode.width() - pWidth.get(0)) / 2,
                    (vidmode.height() - pHeight.get(0)) / 2
            );
        }

        glfwMakeContextCurrent(window);
        glfwSwapInterval(1);
        glfwShowWindow(window);
    }

    private void loop() {
        GL.createCapabilities();

        renderer.init();

        while (!glfwWindowShouldClose(window)) {

            renderer.display();

            glfwSwapBuffers(window);

            glfwPollEvents();
        }
    }

}