package renderer;

import model.Canvas;
import model.Controller;
import lwjglutils.OGLUtils;
import org.lwjgl.glfw.*;

import java.util.ArrayList;
import java.util.List;

import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;

public abstract class AbstractRenderer {

    private int pass;
    protected int width;
    protected int height;
    protected Controller controller;
    protected Canvas canvas;
    protected List<Canvas> canvases = new ArrayList<>();

    public AbstractRenderer(Controller controller) {
        this.width = 600;
        this.height = 400;
        this.controller = controller;
    }

    public void init(){
        OGLUtils.printOGLparameters();
        OGLUtils.printLWJLparameters();
        OGLUtils.printJAVAparameters();
        OGLUtils.shaderCheck();
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    }

    public void display() {
        glViewport(0, 0, width, height);

        pass++;
        glClearColor(
                (float) (Math.sin(pass / 100.) / 2 + 0.5),
                (float) (Math.cos(pass / 200.) / 2 + 0.5),
                (float) (Math.sin(pass / 300.) / 2 + 0.5),
                0.0f
        );
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    protected GLFWKeyCallback glfwKeyCallback = new GLFWKeyCallback() {
        @Override
        public void invoke(long window, int key, int scancode, int action, int mods) {
            if (action == GLFW_RELEASE) {
                controller.keyReleased(key);
            }
            if (action == GLFW_PRESS) {
                controller.keyPressed(key);
                canvas.keyPressed(key);
            }
        }
    };

    protected GLFWWindowSizeCallback glfwWindowSizeCallback = new GLFWWindowSizeCallback() {
        @Override
        public void invoke(long window, int w, int h) {
            if (w > 0 && h > 0) {
                width = w;
                height = h;
            }
        }
    };
    protected GLFWCharCallback glfwCharCallback = new GLFWCharCallback() {
        @Override
        public void invoke(long window, int codepoint) {
            canvas.keyTyped(codepoint);
        }
    };

    protected GLFWMouseButtonCallback glfwMouseButtonCallback = new GLFWMouseButtonCallback() {

        @Override
        public void invoke(long window, int button, int action, int mods) {

        }

    };

    protected GLFWCursorPosCallback glfwCursorPosCallback = new GLFWCursorPosCallback() {
        @Override
        public void invoke(long window, double x, double y) {
            controller.changeMousePos(x, y);
        }
    };

    protected GLFWScrollCallback glfwScrollCallback = new GLFWScrollCallback() {
        @Override
        public void invoke(long window, double dx, double dy) {
        }
    };

    public GLFWKeyCallback getGlfwKeyCallback() {
        return glfwKeyCallback;
    }

    public GLFWWindowSizeCallback getGlfwWindowSizeCallback() {
        return glfwWindowSizeCallback;
    }

    public GLFWMouseButtonCallback getGlfwMouseButtonCallback() {
        return glfwMouseButtonCallback;
    }

    public GLFWCursorPosCallback getGlfwCursorPosCallback() {
        return glfwCursorPosCallback;
    }

    public GLFWScrollCallback getGlfwScrollCallback() {
        return glfwScrollCallback;
    }
    public GLFWCharCallback getGlfwCharCallback() {
        return glfwCharCallback;
    }

    public void dispose() {

    }
}
