package ui.components;

import lwjglutils.OGLTextRenderer;
import model.Canvas;
import ui.utils.UITextAlign;
import ui.utils.UITextProvider;

import java.awt.*;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;

public class UILabel extends UIComponent implements UITextProvider {
    private String text;
    private OGLTextRenderer textRenderer;
    private UITextAlign align;
    private int offsetX, offsetY;
    private int canvasPosX, canvasPosY;


    /**
     * @param posX   position x in ndc
     * @param posY   position y in ndc
     * @param text   text to be displayed
     * @param canvas Canvas
     * @param align  alignment of the text
     * @param width  width of the label
     * @param height height of the label
     */
    public UILabel(float width, float height, float posX, float posY, String text, Canvas canvas, UITextAlign align) {
        super(width, height, posX, posY, canvas);
        this.textRenderer = canvas.getTextRenderer().clone();
        this.align = align;
        canvasPosX = Math.round(((x + 1) * canvas.getWidth()) / 2);
        canvasPosY = Math.round((((y * -1) + 1) * canvas.getHeight()) / 2);
        setText(text);
    }

    public void setText(String text) {
        this.text = text;
        if (align == UITextAlign.CENTER) {
            offsetX = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getWidth() / 2);
            offsetY = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getHeight() / 2);
        } else if (align == UITextAlign.CENTER_X) {
            offsetX = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getWidth() / 2);
        } else if (align == UITextAlign.CENTER_Y) {
            offsetY = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getHeight() / 2);
        }

    }

    public String getText() {
        return text;
    }

    public void setTextColor(Color color) {
        if (color == null)
            textRenderer.setColor(Color.WHITE);
        else {
            if (color.equals(Color.BLACK))
                textRenderer.setColor(new Color(0.01f, 0.01f, 0.01f));
            else
                textRenderer.setColor(color);
        }

    }

    public void setBackgroundColor(Color color) {
        textRenderer.setBackgroundColor(color);
    }

    public void invert() {
        textRenderer.invert();
    }


    public void setTextRotation(double angle) {
        textRenderer.setRotationAngle(angle);
    }

    @Override
    public void render() {

    }

    @Override
    public void renderText() {
        if (align == UITextAlign.CENTER)
            textRenderer.addStr2D(canvasPosX - offsetX, canvasPosY + offsetY, text);
        else if (align == UITextAlign.CENTER_X)
            textRenderer.addStr2D(canvasPosX - offsetX, canvasPosY, text);
        else if (align == UITextAlign.CENTER_Y)
            textRenderer.addStr2D(canvasPosX, canvasPosY + offsetY, text);
        else
            textRenderer.addStr2D(canvasPosX, canvasPosY, text);
    }

    @Override
    public void resize(int width, int height) {
        canvasPosX = Math.round(((x + 1) * width) / 2);
        canvasPosY = Math.round((((y * -1) + 1) * height) / 2);
        textRenderer.resize(width, height);
    }

    public UITextAlign getAlign() {
        return align;
    }

    public void setAlign(UITextAlign align) {
        this.align = align;
    }
    public float getTextWidth() {
        int textWidth = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getWidth());
        return (textWidth*2f) / canvas.getWidth();
    }
    public float getTextWidth(String text) {
        int textWidth = (int) Math.round(textRenderer.getFont().getStringBounds(text, new FontRenderContext(new AffineTransform(), true, true)).getWidth());
        return (textWidth*2f) / canvas.getWidth();
    }
}
