package ui.components;

import lwjglutils.OGLBuffers;
import lwjglutils.ShaderUtils;
import model.Canvas;

import java.awt.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL20.glUniform3f;

public class UIPlaceholder extends UIComponent {
    private float r,g,b;
    private boolean autoResize = true;
    private OGLBuffers buffers;
    private int shader;
    private final int colorLoc;

    public UIPlaceholder(float width, float height, float posX, float posY, Color color, Canvas canvas) {
        super(width, height, posX, posY, canvas);
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;
        shader = ShaderUtils.loadProgram("/Button.vert",
                "/Placeholder.frag",
                null,null,null,null);
        glUseProgram(this.shader);
        colorLoc = glGetUniformLocation(shader, "color");

        createBuffers();
    }

    @Override
    public void render() {
        glUseProgram(this.shader);
        glUniform3f(colorLoc, r, g, b);
        buffers.draw(GL_TRIANGLES, shader);
        glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, 0);

    }
    void createBuffers() {
        float[] vertexBufferData = {
                x, y, 	r, g, b,
                x+ componentWidth,  y,		r, g, b,
                x,  y+componentHeight,		r, g, b,
                x+ componentWidth,  y+componentHeight,		r, g, b
        };
        int[] indexBufferData = { 0, 1, 2, 2, 1, 3};

        OGLBuffers.Attrib[] attributes = {
                new OGLBuffers.Attrib("inPosition", 2), // 2 floats
                new OGLBuffers.Attrib("inColor", 3) // 3 floats
        };
        buffers = new OGLBuffers(vertexBufferData, attributes,
                indexBufferData);

    }

    @Override
    public void resize(int width, int height) {
        float aspectRatio = (float)width/(float)height;
        correctedHeight = this.componentHeight * aspectRatio;
        correctedWidth = this.componentWidth * aspectRatio;

        createBuffers();
    }

    public Color getColor() {
        return new Color(r, g, b);
    }

    public void setColor(Color color) {
        r = color.getRed()/255f;
        g = color.getGreen()/255f;
        b = color.getBlue()/255f;
    }
     public void setAutoResize(boolean autoResize) {
         this.autoResize = autoResize;
     }
     public boolean isAutoResize() {
         return autoResize;
     }
}
